package com.tzld.piaoquan.recommend.server.service.score;


import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.common.base.RequestContext;
import com.tzld.piaoquan.recommend.server.common.base.UserFeature;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import com.tzld.piaoquan.recommend.server.service.score.feature.GBDTFeatureExtractorBase;
import com.tzld.piaoquan.recommend.server.service.score.model.GBDTModel;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.concurrent.*;

public class VlogShareGBDTScorer extends BaseGBDTModelScorer {

    private final static int CORE_POOL_SIZE = 64;
    private final static int TIME_OUT = 150;
    private final static Logger LOGGER = LoggerFactory.getLogger(VlogShareGBDTScorer.class);
    private final static ExecutorService executorService = Executors.newFixedThreadPool(CORE_POOL_SIZE);
    private static final String VIMAGE_VIDEO_REC_TAG = "小视频";

    public VlogShareGBDTScorer(ScorerConfigInfo configInfo) {
        super(configInfo);
    }


    @Override
    public List<RankItem> scoring(final ScoreParam param,
                                  final UserFeature userFeature,
                                  final List<RankItem> rankItems) {
        long startTime = System.currentTimeMillis();

        GBDTModel model = (GBDTModel) this.getModel();
        if (model == null) {
            LOGGER.error("not found model");
            return rankItems;
        }

        // 多Rank的rank打分
        multipleGBDTScore(rankItems, model, userFeature, param);
        Collections.sort(rankItems);
        LOGGER.debug("dwelltime ranker excute time: [{}]", System.currentTimeMillis() - startTime);
        return rankItems;
    }

    /**
     * 分数转换
     *
     * @param oldScore
     * @param dwelltimeOrCompletion
     * @param duration
     * @param itemId
     * @return
     */
    private double predictMultipleGBDTScore(final double oldScore, final double dwelltimeOrCompletion,
                                            final double duration, final String itemId,
                                            final RecommendRequest requestData, final UserFeature user) {
        double ctrScalePower = 0;
        double dwelltimeScalePower = 0;

        double ctrScore = oldScore;
        if (oldScore <= 0.0) {
            ctrScore = 0.01;
        }
        double newScore = Math.pow(ctrScore, ctrScalePower) * Math.pow(dwelltimeOrCompletion, dwelltimeScalePower);
        //值越大压制越弱
        double alpha = 1.0;
        double suppressRatio = (1 + alpha) / (Math.max(duration, 300.0) / 300 + alpha);
        newScore = newScore * suppressRatio;
        LOGGER.debug("Expected dwell time score [{}]: {} ,{}, {}--> {}", new Object[]{itemId, oldScore, dwelltimeOrCompletion, duration, newScore});
        return newScore;
    }

    private void gbdtBaseScore(final RankItem item,
                               final GBDTModel model,
                               final RequestContext requestContext,
                               final Map<String, Double> userFeatures,
                               final Map<String, Double> contextFeatures,
                               final UserFeature user) {
        try {

            Map<String, Double> articleFeatures;
            double freshness;
            Map<String, Double> features;
            features = GBDTFeatureExtractorBase.extractUserFeatures(user);

            if (model != null) {
                Map<String, Double> featureScoreMap = new HashMap<String, Double>();
                double pro = model.score(features, featureScoreMap);
                featureScoreMap.put("TOTAL_SCORE", pro);
                pro = pro > 1 ? pro : 1;
                Double rankScore = item.getScore();
                LOGGER.debug("xgb score = {}, lgb score = {}", rankScore, pro);
                double lgbWeight = 1.0;
                pro = (rankScore + lgbWeight * pro) / (1.0 + lgbWeight);
                item.setScore(pro);
            }

        } catch (Exception e) {
            LOGGER.error("Exception {},{}", requestContext, ExceptionUtils.getFullStackTrace(e));
        }
    }


    public void multipleGBDTScore(final List<RankItem> items,
                                  final GBDTModel model,
                                  final UserFeature user,
                                  final ScoreParam param) {
        final int size = items.size();
        if (size == 0) {
            return;
        }

        final Map<String, Double> userFeatures;
        final Map<String, Double> contextFeatures;

        // context feature
        contextFeatures = GBDTFeatureExtractorBase.extractRequestContextFeatures(param);
        // user feature
        userFeatures = GBDTFeatureExtractorBase.extractUserFeatures(user);

        // score item
        List<Callable<Object>> callables = new ArrayList<Callable<Object>>();
        for (int index = 0; index < items.size(); index++) {
            final int fIndex = index;
            callables.add(new Callable<Object>() {
                              @Override
                              public Object call() throws Exception {
                                  try {
                                      gbdtBaseScore(items.get(fIndex), model, param.getRequestContext(), userFeatures, contextFeatures,
                                              user);
                                  } catch (Exception e) {
                                      LOGGER.error("dwelltime exception: [{}] [{}]", items.get(fIndex), ExceptionUtils.getFullStackTrace(e));
                                  }
                                  return new Object();
                              }
                          }
            );
        }

        List<Future<Object>> futures = null;
        try {
            futures = executorService.invokeAll(callables, TIME_OUT, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            LOGGER.error("execute invoke fail: {}", ExceptionUtils.getFullStackTrace(e));
        }

        //等待所有请求的结果返回, 超时也返回
        if (futures != null) {
            for (Future<Object> future : futures) {
                try {
                    if (future != null && future.isDone() && !future.isCancelled() && future.get() != null) {
                    } else {
                        LOGGER.debug("Canceled Dwelltime Score {}", param.getRequestContext());
                    }
                } catch (Exception e) {
                    LOGGER.error("InterruptedException {},{}", ExceptionUtils.getFullStackTrace(e));
                }
            }
        }
    }
}

